//
//	Create a svg file from files generated by freefem++
//	--------------------------------------------------------
//	To compile
//	g++ -lboost_regex -o FreeFem2SVG FreeFem2SVG.cpp
//	--------------------------------------------------------
//	To use
//	FreeFem2SVG [-mesh=%d -width=%d] output.svg file.msh [u]
//
//	mesh=mesh stroke width (0=no mesh plot, default=1)
//	width=width of the SVG minus 2*marging (height is computed automaticaly, defaultwidth:600).
//	output.svg	= output file (vtk format)
//	file1.msh	= mesh file (msh format)
//	u		= P1 finite element

// TO DO:
//	- Extensions: Making animations
//	FreeFem2SVG [-mesh=%d] output.svg file.msh [u]* [file.msh [u]*]*
//	%2=mesh width (0=no mesh plot, default=1)
//	If only one mesh/finite element, standard ploting
//	If more than one argument => animation
//
//	For animations, we have to determine the max and min of the finite elements and of all the meshes
//	This should be done by a preprocessing step
//
//	- Add options for color different color types: Basically colormax and colormin
//	- define a triangle, edge and vertex class (would be more natural)
//	- change the path definition, so to change the fill style in the path itself (if possible ? it seems verbose to me)
//	- change the gradient for local bounding box (better for translation, rotation,....)
//	- use fastformat instead of sprintf ?
//	- treat exceptions consistently

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <string>
#include <boost/regex.hpp> // I tried <regex> but it didn't seem to work properly.
using namespace std;

class mesh
{
	private:
	double min_x_,min_y_,max_x_,max_y_;
	int *Tv,*ev;		// Triangles vertices, Edges vertices
	int *Tlabel,*elabel,*vlabel;	// Triangle, edges ans=d vertices labels
	public:
	int stroke_width;	// stroke width of the mesh (0=no stroke), default=1
	const char* FileName;	// Name of the mesh file
	double *x,*y;		// vertices coordinates
	int nv,nt,ne;		// nb vertives, triangles, edges
	int operator () (const int i) const {return  Tv[i]-1;};
	double min_x() const {return min_x_;};
	double min_y() const {return min_y_;};
	double max_x() const {return max_x_;};
	double max_y() const {return max_y_;};
	mesh(const char* FileName_);
};

class EFP1
{
	
	public:
	mesh const * const Th;
	private:
	double *u;
	public:
	double operator () (const int i) const {return u[i];}
	EFP1 (const mesh Th,const char* Filename);
};

class color
{
	private:
		double maxVal,minVal;
	public:
	string val(const double intensity)
	{
		char res[19];	// But it would be better to use fastformat and strings
		int Intensity;
		{
			double dVal=maxVal-minVal;
			double ValScale=0;		
			if(dVal!=0) ValScale=255/dVal;
			if(dVal!=minVal)
			{
				Intensity=(int) ((intensity-minVal)*ValScale);
			}
		}
		sprintf(res,"rgb(%d,%d,%d)",Intensity,Intensity,Intensity);
		//fastformat::fmt(res, "rgb({0},{0},{0})",Intensity);	
		return res;	
	}
	color(const EFP1& u)
	{
		minVal=u(0);maxVal=minVal;
		for(int i=1;i<(u.Th)->nv;i++)
		{
			minVal=min(minVal,u(i));
			maxVal=max(maxVal,u(i));
		}
	}
};


class svg
{
private:
	const string FileName;
	double size_x,size_y,margin;
public:
	void add(const mesh&,const int&);
	void add(const EFP1&,const int&);
	svg(const string,const double, const double,const double);
};


mesh::mesh(const char* FileName_):FileName(FileName_)
{
	boost::regex e ("(.*)\\.msh");
	boost::match_results<std::string::const_iterator>  sm;
	if (!(boost::regex_match ((string) FileName,e)))
	{
		cout<<FileName<<" should be of format .msh"<<endl;
		exit(1);
	}
	try
	{
		ifstream MeshFile(FileName);
		if(!MeshFile.good()) throw 1;
		MeshFile>>nv;
		MeshFile>>nt;
		MeshFile>>ne;
		double* Z;
		x=new double[nv];	
		y=new double[nv];
		vlabel=new int[nv];
		MeshFile>>x[0]>>y[0]>>vlabel[0];
		min_x_=x[0];max_x_=x[0];
		min_y_=y[0];max_y_=y[0];
		for(int i=1;i<nv;i++)
		{
			MeshFile>>x[i]>>y[i]>>vlabel[i];
			min_x_=min(x[i],min_x_);max_x_=max(x[i],max_x_);
			min_y_=min(y[i],min_y_);max_y_=max(y[i],max_y_);
		}
		
		Tv=new int[nt*3];Tlabel=new int[nt];
		for(int i=0;i<nt;i++)
		{
			for (int k=0;k<3;k++) MeshFile>>Tv[3*i+k];
			MeshFile>>Tlabel[i]; 
		}
	
		ev=new int[2*ne];elabel=new int[2*ne];
		for(int i=0;i<ne;i++)
		{
			for (int k=0;k<2;k++) MeshFile>>ev[2*i+k];
			MeshFile>>elabel[i]; 
		}
		MeshFile.close();
	}
	catch (bad_alloc&)
	{
		cout<<"Wrong format of "<<FileName<<endl;
		exit(1);
	}
	catch (int)
	{
		cout<<"File "<<FileName<<" does not exist"<<endl;
		exit(1);
	}
};

class Wrong {} wrong; // Not a wonderful way to take excpetion into account... the treatment of exception should be made more consistent.
EFP1::EFP1 (const mesh Th_,const char* FileName):Th(&Th_),u(new double[Th->nv])
{
	try
	{
		ifstream FEFile(FileName);
		if(!(FEFile.good())) throw (FileName);
		int ndof; FEFile>>ndof;
		if(ndof!=Th->nv) throw (ndof);
		else
		for(int i=0;i<Th->nv;i++) if(FEFile.eof()) {throw (wrong);} else FEFile>>u[i];
		FEFile.close();
	}
	catch (bad_alloc&)
	{
		cout<<"Wrong format of "<<FileName<<endl;
		exit(1);
	}
	catch (Wrong&)
	{
		cout<<"Wrong format of "<<FileName<<endl;
		exit(1);
	}
	catch (const char* FileName)
	{
		cout<<"File "<<FileName<<" does not exist"<<endl;
		exit(1);
	}
	catch (int& ndof)
	{
		cout<<"Wrong dimension of "<<FileName<<" with respect to the mesh "<< Th->FileName<<": "<<ndof<<" dof "<<"instead of "<<Th->nv<<endl;exit(1);
		exit(1);
	}
};

void svg::add(const mesh& Th,const int& meshStroke=1)
{
		fstream SVGFile;
		SVGFile.open(FileName.c_str(),fstream::in | fstream::out);
		SVGFile.seekp(-7,ios::end);
		SVGFile<<"<path style=\"fill:none;stroke:black; stroke-width:"<<meshStroke<<";  stroke-linejoin:round;\" d=\"";
		double scale=size_x/(Th.max_x()-Th.min_x());
		for(int i=0;i<Th.nt;i++)
		{
			SVGFile<<"M ";
			for(int j=0;j<3;j++) SVGFile<<(Th.x[Th(3*i+j)]-Th.min_x())*scale+margin<<","<<(Th.max_y()-Th.y[Th(3*i+j)])*scale+margin<<" ";
			SVGFile<<"Z ";
		}
		SVGFile<<"\"></path>"<<endl;
		SVGFile<<"</svg>"<<endl;
		SVGFile.close();

}
void svg::add(const EFP1& u,const int& meshStroke=1)
{
	color MyColor(u);
	const mesh& Th=*(u.Th);
	fstream SVGFile;
	SVGFile.open(FileName.c_str(),fstream::in | fstream::out);
	SVGFile.seekp(-7,ios::end);
	double scale=size_x/(Th.max_x()-Th.min_x());
	// definition of the gradients
	SVGFile<<"<defs>"<<endl;
	for(int i=0;i<Th.nt;i++)
	{
		// u restricted to a triangle K is P1,
		// that is u=du.x+c
		// We set imin=argmin du.x[i], imax=argmax du.x[i]
		// We then set x1=x[imin], y1=y[imin]
		// We then set x2=x[imin], y2=y[imin]
		// and the color is given by the values of u at imin and imax
		
		// Computation of the gradient.
		double dux=0,duy=0;
		for(int j=0;j<3;j++)
		{
			int k=Th(3*i+j),k1=Th(3*i+((j+1)%3)),k2=Th(3*i+((j+2)%3));
			double nx=-(Th.y[k2]-Th.y[k1]),ny=Th.x[k2]-Th.x[k1];
			double rescale=nx*(Th.x[k]-Th.x[k1])+ny*(Th.y[k]-Th.y[k1]);
			nx=nx/rescale;ny=ny/rescale;
			dux+=u(k)*nx;duy+=u(k)*ny;
		}

		// Computation of extremal points
		int jmin=0,jmax=0;	// in 0...2
		double dumin,dumax;
		{int k=Th(3*i); dumin=Th.x[k]*dux+Th.y[k]*duy; dumax=dumin;}
		for(int j=1;j<3;j++)
		{
			int k=Th(3*i+j);
			double du=Th.x[k]*dux+Th.y[k]*duy;
			if(du<dumin){jmin=j;dumin=du;}
			if(du>dumax){jmax=j;dumax=du;}
		}
		int kmin=Th(3*i+jmin),kmax=Th(3*i+jmax);
		// x1 argmin u
		// x2-x1=lambda*du
		// (x2-x1).du=(armax u-armin u).du
		// lambda=((armax u-armin u).du)/|du|^2
		double normdu=dux*dux+duy*duy;
		double x1,y1,x2,y2;
		if(normdu!=0)
		{	
			double lambda=((Th.x[kmax]-Th.x[kmin])*dux+(Th.y[kmax]-Th.y[kmin])*duy)/normdu;
			double X1=Th.x[kmin],Y1=Th.y[kmin];		
			double X2=X1+lambda*dux,Y2=Y1+lambda*duy;
			x1=(X1-Th.min_x())*scale+margin;y1=(Th.max_y()-Y1)*scale+margin; 
			x2=(X2-Th.min_x())*scale+margin;y2=(Th.max_y()-Y2)*scale+margin;
		} 
		else
		{
			double X1=Th.x[kmin],Y1=Th.y[kmin];		
			x1=(X1-Th.min_x())*scale;y1=(Th.max_y()-Y1)*scale;
			x2=x1;y2=y1; 
		} 
		SVGFile<<"<linearGradient id=\"lg"<<i<<"\" x1=\""<<x1<<"\" y1=\""<<y1<<"\" x2=\""<<x2<<"\" y2=\""<<y2<<"\" gradientUnits=\"userSpaceOnUse\">";
		SVGFile<<"<stop style=\"stop-color:"<<MyColor.val(u(kmin))<<";\" offset=\"0\" />"; 
		SVGFile<<"<stop style=\"stop-color:"<<MyColor.val(u(kmax))<<";\" offset=\"1\" />";
		SVGFile<<"</linearGradient>\n";
	}
	SVGFile<<"</defs>"<<endl;
	// lets draw the graph	
	for(int i=0;i<Th.nt;i++)
	{
		if(meshStroke==0)
			{SVGFile<<"<path style=\"fill:url(#lg"<<i<<");stroke:url(#lg"<<i<<"); stroke-width:1;  stroke-linejoin:round;\" d=\"";}// Hide mesh
		else	{SVGFile<<"<path style=\"fill:url(#lg"<<i<<");stroke:black; stroke-width:"<<meshStroke<<";  stroke-linejoin:round;\" d=\""; }// Show mesh
		SVGFile<<"M ";
		for(int j=0;j<3;j++) SVGFile<<(Th.x[Th(3*i+j)]-Th.min_x())*scale+margin<<","<<(Th.max_y()-Th.y[Th(3*i+j)])*scale+margin<<" ";
		SVGFile<<"Z ";
		SVGFile<<"\"></path>"<<endl;
	}
	SVGFile<<"</svg>"<<endl;
	SVGFile.close();
}

svg::svg(const string FileName_,const double size_x_=600, const double size_y_=600,const double margin_=0):FileName(FileName_),size_x(size_x_),size_y(size_y_),margin(margin_) 
{
	ofstream SVGFile(FileName.c_str());
	SVGFile<<"<svg xmlns=\"http://www.w3.org/2000/svg\" width=\""<<(size_x+2*margin)<<"\" height=\""<<(size_y+2*margin)<<"\">"<<endl;
	SVGFile<<"</svg>"<<endl;
	SVGFile.close();
};

void ArgumentsError()
{
		cout<<"FreeFem2SVG -- version 0 --- 21/03/2014"<<endl;
		cout<<"-- Use --------"<<endl;
		cout<<"     FreeFem2VVG.exe output.svg file1.msh [u1]"<<endl;
		cout<<"            output.svg          = Output file (svg format)"<<endl;
		cout<<"            file1.msh          = Mesh file (msh format)"<<endl;
		cout<<"            u1                  = P1 Finite Element file"<<endl;
		cout<<endl;
		cout<<"Incorrect Number of arguments or format"<<endl;
		exit(1);
}
int main(int argc,char* argv[])
{
	
//		argc = nb of elements in argv
//		argv[1..(argc-1)] = list of the arguments
//		argv[0] is the command line (minus the argument)

	// Default parameter
	int meshStroke=1;	// default mesh stroke size
	double width=600;	// default width of the SVG output
	int margin=-1;		// margin... default is set to meshStroke/2

	// Look at options
	int currentArgument=1;
	if(argc==1) ArgumentsError();
	while (argv[currentArgument][0]=='-')
	{
		// It is an option
		bool found=false;	// can it be found ?	
		{//mesh stroke width
			boost::regex e ("-mesh=([0-9]*)");
			boost::match_results<std::string::const_iterator>  sm;
			if(boost::regex_match ((string) argv[currentArgument],sm,e))
			{	
				string theMeshStroke=sm[1];
				meshStroke=atoi(theMeshStroke.c_str());
				currentArgument++;found=true;
			}
		}//mesh stroke width
		{//width of the SVG
			boost::regex e ("-width=([0-9\\.]*)");
			boost::match_results<std::string::const_iterator>  sm;
			if(boost::regex_match ((string) argv[currentArgument],sm,e))
			{	
				string theWidth=sm[1];
				width=atoi(theWidth.c_str());
				cout<<"width="<<width<<endl;
				currentArgument++;found=true;
			}
		}//width of the SVG
		{//margin of the SVG
			boost::regex e ("-margin=([0-9\\.]*)");
			boost::match_results<std::string::const_iterator>  sm;
			if(boost::regex_match ((string) argv[currentArgument],sm,e))
			{	
				string thePadding=sm[1];
				margin=atoi(thePadding.c_str());
				cout<<"margin="<<margin<<endl;
				currentArgument++;found=true;
			}
		}//margin of the SVG
		if(!found)
		{
			cout<<"Option "<<argv[currentArgument]<<" unknown"<<endl;exit(1);
		}
	}
	if (margin==-1) margin=meshStroke/2.;	

	// Check nb of files
	if ((argc!=2+currentArgument) and (argc!=3+currentArgument)) ArgumentsError();
	
	string output=argv[currentArgument++];
	mesh Th(argv[currentArgument++]);	// Read mesh
	
 	// create SVG FIle
	double height=600;
	if(Th.max_x()!=Th.min_x()) height=(Th.max_y()-Th.min_y())/(Th.max_x()-Th.min_x())*width;
	svg SVG(output,width,height,margin);	
	if(argc>currentArgument) 
	{
		cout<<"Loading "<<argv[currentArgument]<<endl;
		EFP1 u(Th,argv[currentArgument++]);	// Read FE P1
		SVG.add(u,meshStroke);		// add finite element:w
	}
	else SVG.add(Th,meshStroke);		// add the mesh
}
